@@ -1,5 +1,5 @@
 /* Initializes a new PO file.
-   Copyright (C) 2001-2003 Free Software Foundation, Inc.
+   Copyright (C) 2001-2006 Free Software Foundation, Inc.
    Written by Bruno Haible <haible@clisp.cons.org>, 2001.
 
    This program is free software; you can redistribute it and/or modify
@@ -14,7 +14,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
 
 
 #ifdef HAVE_CONFIG_H
@@ -37,33 +37,13 @@
 # include <pwd.h>
 #endif
 
-#if HAVE_UNISTD_H
-# include <unistd.h>
-#endif
+#include <unistd.h>
 
 #if HAVE_DIRENT_H
 # include <dirent.h>
-#else
-# define dirent direct
-# if HAVE_SYS_NDIR_H
-#  include <sys/ndir.h>
-# endif
-# if HAVE_SYS_DIR_H
-#  include <sys/dir.h>
-# endif
-# if HAVE_NDIR_H
-#  include <ndir.h>
-# endif
 #endif
 
-#if CLOSEDIR_VOID
-/* Fake a return value. */
-# define CLOSEDIR(d) (closedir (d), 0)
-#else
-# define CLOSEDIR(d) closedir (d)
-#endif
-
-#if HAVE_DIRENT_H || HAVE_NDIR_H || HAVE_SYS_DIR_H || HAVE_SYS_NDIR_H
+#if HAVE_DIRENT_H
 # define HAVE_DIR 1
 #else
 # define HAVE_DIR 0
@@ -76,25 +56,36 @@
 #include "relocatable.h"
 #include "basename.h"
 #include "strpbrk.h"
-#include "strstr.h"
-#include "strcase.h"
+#include "c-strstr.h"
+#include "c-strcase.h"
 #include "message.h"
+#include "read-catalog.h"
 #include "read-po.h"
+#include "read-properties.h"
+#include "read-stringtable.h"
+#include "write-catalog.h"
 #include "write-po.h"
+#include "write-properties.h"
+#include "write-stringtable.h"
 #include "po-charset.h"
 #include "localcharset.h"
 #include "po-time.h"
 #include "plural-table.h"
+#include "lang-table.h"
 #include "xalloc.h"
+#include "xallocsa.h"
 #include "exit.h"
 #include "pathname.h"
 #include "xerror.h"
+#include "xvasprintf.h"
 #include "msgl-english.h"
+#include "plural-count.h"
 #include "pipe.h"
 #include "wait-process.h"
 #include "getline.h"
 #include "xsetenv.h"
 #include "str-list.h"
+#include "propername.h"
 #include "gettext.h"
 
 #define _(str) gettext (str)
@@ -151,6 +142,7 @@ static const char *catalogname_for_locale (const char *locale);
 static const char *language_of_locale (const char *locale);
 static char *get_field (const char *header, const char *field);
 static msgdomain_list_ty *fill_header (msgdomain_list_ty *mdlp);
+static msgdomain_list_ty *update_msgstr_plurals (msgdomain_list_ty *mdlp);
 
 
 int
@@ -162,6 +154,8 @@ main (int argc, char **argv)
   char *output_file;
   const char *input_file;
   msgdomain_list_ty *result;
+  catalog_input_format_ty input_syntax = &input_format_po;
+  catalog_output_format_ty output_syntax = &output_format_po;
 
   /* Set program name for messages.  */
   set_program_name (argv[0]);
@@ -174,6 +168,7 @@ main (int argc, char **argv)
 
   /* Set the text message domain.  */
   bindtextdomain (PACKAGE, relocate (LOCALEDIR));
+  bindtextdomain ("bison-runtime", relocate (BISON_LOCALEDIR));
   textdomain (PACKAGE);
 
   /* Ensure that write errors on stdout are detected.  */
@@ -215,11 +210,11 @@ main (int argc, char **argv)
 	break;
 
       case 'p':
-	message_print_syntax_properties ();
+	output_syntax = &output_format_properties;
 	break;
 
       case 'P':
-	input_syntax = syntax_properties;
+	input_syntax = &input_format_properties;
 	break;
 
       case 'V':
@@ -245,11 +240,11 @@ main (int argc, char **argv)
 	break;
 
       case CHAR_MAX + 3: /* --stringtable-input */
-	input_syntax = syntax_stringtable;
+	input_syntax = &input_format_stringtable;
 	break;
 
       case CHAR_MAX + 4: /* --stringtable-output */
-	message_print_syntax_stringtable ();
+	output_syntax = &output_format_stringtable;
 	break;
 
       default:
@@ -266,8 +261,8 @@ main (int argc, char **argv)
 This is free software; see the source for copying conditions.  There is NO\n\
 warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
 "),
-	      "2001-2003");
-      printf (_("Written by %s.\n"), "Bruno Haible");
+	      "2001-2006");
+      printf (_("Written by %s.\n"), proper_name ("Bruno Haible"));
       exit (EXIT_SUCCESS);
     }
 
@@ -308,11 +303,7 @@ file.  This is necessary so you can test your translations.\n")));
   /* Default output file name is CATALOGNAME.po.  */
   if (output_file == NULL)
     {
-      size_t cnlen = strlen (catalogname);
-
-      output_file = (char *) xmalloc (cnlen + 3 + 1);
-      memcpy (output_file, catalogname, cnlen);
-      memcpy (output_file + cnlen, ".po", 3 + 1);
+      output_file = xasprintf ("%s.po", catalogname);
 
       /* But don't overwrite existing PO files.  */
       if (access (output_file, F_OK) == 0)
@@ -328,7 +319,7 @@ the output .po file through the --output-file option.\n"),
     }
 
   /* Read input file.  */
-  result = read_po_file (input_file);
+  result = read_catalog_file (input_file, input_syntax);
 
   /* Fill the header entry.  */
   result = fill_header (result);
@@ -336,9 +327,11 @@ the output .po file through the --output-file option.\n"),
   /* Initialize translations.  */
   if (strcmp (language, "en") == 0)
     result = msgdomain_list_english (result);
+  else
+    result = update_msgstr_plurals (result);
 
   /* Write the modified message list out.  */
-  msgdomain_list_print (result, output_file, true, false);
+  msgdomain_list_print (result, output_file, output_syntax, true, false);
 
   if (!no_translator)
     fprintf (stderr, "\n");
@@ -465,7 +458,7 @@ Please specify the input .pot file through the --input option.\n")));
 	  else
 	    break;
 	}
-      if (CLOSEDIR (dirp))
+      if (closedir (dirp))
 	error (EXIT_FAILURE, errno, _("error reading current directory"));
 
       if (found != NULL)
@@ -494,25 +487,46 @@ catalogname_for_locale (const char *locale)
 {
   static const char *locales_with_principal_territory[] = {
 		/* Language	Main territory */
+    "ace_ID",	/* Achinese	Indonesia */
     "af_ZA",	/* Afrikaans	South Africa */
+    "ak_GH",	/* Akan		Ghana */
     "am_ET",	/* Amharic	Ethiopia */
     "an_ES",	/* Aragonese	Spain */
+    "ang_GB",	/* Old English	Britain */
     "as_IN",	/* Assamese	India */
+    "av_RU",	/* Avaric	Russia */
+    "awa_IN",	/* Awadhi	India */
     "az_AZ",	/* Azerbaijani	Azerbaijan */
+    "bad_CF",	/* Banda	Central African Republic */
+    "ban_ID",	/* Balinese	Indonesia */
     "be_BY",	/* Belarusian	Belarus */
+    "bem_ZM",	/* Bemba	Zambia */
     "bg_BG",	/* Bulgarian	Bulgaria */
+    "bho_IN",	/* Bhojpuri	India */
+    "bik_PH",	/* Bikol	Philippines */
+    "bin_NG",	/* Bini		Nigeria */
+    "bm_ML",	/* Bambara	Mali */
     "bn_IN",	/* Bengali	India */
     "bo_CN",	/* Tibetan	China */
     "br_FR",	/* Breton	France */
     "bs_BA",	/* Bosnian	Bosnia */
+    "btk_ID",	/* Batak	Indonesia */
+    "bug_ID",	/* Buginese	Indonesia */
     "ca_ES",	/* Catalan	Spain */
     "ce_RU",	/* Chechen	Russia */
+    "ceb_PH",	/* Cebuano	Philippines */
     "co_FR",	/* Corsican	France */
+    "cr_CA",	/* Cree		Canada */
     "cs_CZ",	/* Czech	Czech Republic */
+    "csb_PL",	/* Kashubian	Poland */
     "cy_GB",	/* Welsh	Britain */
     "da_DK",	/* Danish	Denmark */
     "de_DE",	/* German	Germany */
+    "din_SD",	/* Dinka	Sudan */
+    "doi_IN",	/* Dogri	India */
+    "dv_MV",	/* Divehi	Maldives */
     "dz_BT",	/* Dzongkha	Bhutan */
+    "ee_GH",	/* Éwé		Ghana */
     "el_GR",	/* Greek	Greece */
     /* Don't put "en_GB" or "en_US" here.  That would be asking for fruitless
        political discussion.  */
@@ -520,84 +534,145 @@ catalogname_for_locale (const char *locale)
     "et_EE",	/* Estonian	Estonia */
     "fa_IR",	/* Persian	Iran */
     "fi_FI",	/* Finnish	Finland */
+    "fil_PH",	/* Filipino	Philippines */
     "fj_FJ",	/* Fijian	Fiji */
     "fo_FO",	/* Faroese	Faeroe Islands */
+    "fon_BJ",	/* Fon		Benin */
     "fr_FR",	/* French	France */
+    "fy_NL",	/* Western Frisian	Netherlands */
     "ga_IE",	/* Irish	Ireland */
     "gd_GB",	/* Scots	Britain */
+    "gon_IN",	/* Gondi	India */
+    "gsw_CH",	/* Swiss German	Switzerland */
     "gu_IN",	/* Gujarati	India */
     "he_IL",	/* Hebrew	Israel */
     "hi_IN",	/* Hindi	India */
+    "hil_PH",	/* Hiligaynon	Philippines */
     "hr_HR",	/* Croatian	Croatia */
     "ht_HT",	/* Haitian	Haiti */
     "hu_HU",	/* Hungarian	Hungary */
     "hy_AM",	/* Armenian	Armenia */
     "id_ID",	/* Indonesian	Indonesia */
+    "ig_NG",	/* Igbo		Nigeria */
     "ii_CN",	/* Sichuan Yi	China */
+    "ilo_PH",	/* Iloko	Philippines */
     "is_IS",	/* Icelandic	Iceland */
     "it_IT",	/* Italian	Italy */
     "ja_JP",	/* Japanese	Japan */
+    "jab_NG",	/* Hyam		Nigeria */
     "jv_ID",	/* Javanese	Indonesia */
     "ka_GE",	/* Georgian	Georgia */
+    "kab_DZ",	/* Kabyle	Algeria */
+    "kaj_NG",	/* Jju		Nigeria */
+    "kam_KE",	/* Kamba	Kenya */
+    "kmb_AO",	/* Kimbundu	Angola */
+    "kcg_NG",	/* Tyap		Nigeria */
+    "kdm_NG",	/* Kagoma	Nigeria */
+    "kg_CD",	/* Kongo	Democratic Republic of Congo */
     "kk_KZ",	/* Kazakh	Kazakhstan */
     "kl_GL",	/* Kalaallisut	Greenland */
     "km_KH",	/* Khmer	Cambodia */
     "kn_IN",	/* Kannada	India */
     "ko_KR",	/* Korean	Korea (South) */
     "kok_IN",	/* Konkani	India */
+    "kr_NG",	/* Kanuri	Nigeria */
+    "kru_IN",	/* Kurukh	India */
+    "lg_UG",	/* Ganda	Uganda */
     "li_BE",	/* Limburgish	Belgium */
     "lo_LA",	/* Laotian	Laos */
     "lt_LT",	/* Lithuanian	Lithuania */
+    "lu_CD",	/* Luba-Katanga	Democratic Republic of Congo */
+    "lua_CD",	/* Luba-Lulua	Democratic Republic of Congo */
+    "luo_KE",	/* Luo		Kenya */
     "lv_LV",	/* Latvian	Latvia */
+    "mad_ID",	/* Madurese	Indonesia */
+    "mag_IN",	/* Magahi	India */
+    "mai_IN",	/* Maithili	India */
+    "mak_ID",	/* Makasar	Indonesia */
+    "man_ML",	/* Mandingo	Mali */
+    "men_SL",	/* Mende	Sierra Leone */
     "mg_MG",	/* Malagasy	Madagascar */
+    "min_ID",	/* Minangkabau	Indonesia */
     "mk_MK",	/* Macedonian	Macedonia */
     "ml_IN",	/* Malayalam	India */
     "mn_MN",	/* Mongolian	Mongolia */
+    "mni_IN",	/* Manipuri	India */
+    "mos_BF",	/* Mossi	Burkina Faso */
     "mr_IN",	/* Marathi	India */
     "ms_MY",	/* Malay	Malaysia */
     "mt_MT",	/* Maltese	Malta */
+    "mwr_IN",	/* Marwari	India */
     "my_MM",	/* Burmese	Myanmar */
-    "mni_IN",	/* Manipuri	India */
     "na_NR",	/* Nauru	Nauru */
+    "nah_MX",	/* Nahuatl	Mexico */
+    "nap_IT",	/* Neapolitan	Italy */
     "nb_NO",	/* Norwegian Bokmål	Norway */
+    "nds_DE",	/* Low Saxon	Germany */
     "ne_NP",	/* Nepali	Nepal */
     "nl_NL",	/* Dutch	Netherlands */
     "nn_NO",	/* Norwegian Nynorsk	Norway */
     "no_NO",	/* Norwegian	Norway */
+    "nr_ZA",	/* South Ndebele	South Africa */
+    "nso_ZA",	/* Northern Sotho	South Africa */
+    "nym_TZ",	/* Nyamwezi	Tanzania */
+    "nyn_UG",	/* Nyankole	Uganda */
     "oc_FR",	/* Occitan	France */
+    "oj_CA",	/* Ojibwa	Canada */
     "or_IN",	/* Oriya	India */
     "pa_IN",	/* Punjabi	India */
+    "pag_PH",	/* Pangasinan	Philippines */
+    "pam_PH",	/* Pampanga	Philippines */
+    "pbb_CO",	/* Páez		Colombia */
     "pl_PL",	/* Polish	Poland */
     "ps_AF",	/* Pashto	Afghanistan */
     "pt_PT",	/* Portuguese	Portugal */
+    "raj_IN",	/* Rajasthani	India */
     "rm_CH",	/* Rhaeto-Roman	Switzerland */
     "rn_BI",	/* Kirundi	Burundi */
     "ro_RO",	/* Romanian	Romania */
     "ru_RU",	/* Russian	Russia */
     "sa_IN",	/* Sanskrit	India */
+    "sas_ID",	/* Sasak	Indonesia */
+    "sat_IN",	/* Santali	India */
     "sc_IT",	/* Sardinian	Italy */
-    "sg_CF",	/* Sango	Central African Rep. */
-    "si_LK",	/* Sinhalese	Sri Lanka */
+    "scn_IT",	/* Sicilian	Italy */
+    "sg_CF",	/* Sango	Central African Republic */
+    "shn_MM",	/* Shan		Myanmar */
+    "si_LK",	/* Sinhala	Sri Lanka */
+    "sid_ET",	/* Sidamo	Ethiopia */
     "sk_SK",	/* Slovak	Slovakia */
     "sl_SI",	/* Slovenian	Slovenia */
     "so_SO",	/* Somali	Somalia */
     "sq_AL",	/* Albanian	Albania */
-    "sr_CS",	/* Serbian	Serbia & Montenegro */
+    "sr_RS",	/* Serbian	Serbia */
     "sr_YU",	/* Serbian	Yugoslavia */
+    "srr_SN",	/* Serer	Senegal */
+    "suk_TZ",	/* Sukuma	Tanzania */
+    "sus_GN",	/* Susu		Guinea */
     "sv_SE",	/* Swedish	Sweden */
     "te_IN",	/* Telugu	India */
+    "tem_SL",	/* Timne	Sierra Leone */
+    "tet_ID",	/* Tetum	Indonesia */
     "tg_TJ",	/* Tajik	Tajikistan */
     "th_TH",	/* Thai		Thailand */
+    "tiv_NG",	/* Tiv		Nigeria */
     "tk_TM",	/* Turkmen	Turkmenistan */
     "tl_PH",	/* Tagalog	Philippines */
     "to_TO",	/* Tonga	Tonga */
     "tr_TR",	/* Turkish	Turkey */
+    "tum_MW",	/* Tumbuka	Malawi */
     "uk_UA",	/* Ukrainian	Ukraine */
+    "umb_AO",	/* Umbundu	Angola */
     "ur_PK",	/* Urdu		Pakistan */
     "uz_UZ",	/* Uzbek	Uzbekistan */
+    "ve_ZA",	/* Venda	South Africa */
     "vi_VN",	/* Vietnamese	Vietnam */
     "wa_BE",	/* Walloon	Belgium */
-    "wen_DE"	/* Sorbian	Germany */
+    "wal_ET",	/* Walamo	Ethiopia */
+    "war_PH",	/* Waray	Philippines */
+    "wen_DE",	/* Sorbian	Germany */
+    "yao_MW",	/* Yao		Malawi */
+    "zap_MX"	/* Zapotec	Mexico */
   };
   const char *dot;
   size_t i;
@@ -609,7 +684,7 @@ catalogname_for_locale (const char *locale)
       const char *codeset_end;
       char *shorter_locale;
 
-      codeset_end = strpbrk (dot + 1, "_@+,");
+      codeset_end = strpbrk (dot + 1, "_@");
       if (codeset_end == NULL)
 	codeset_end = dot + strlen (dot);
 
@@ -649,7 +724,7 @@ language_of_locale (const char *locale)
 {
   const char *language_end;
 
-  language_end = strpbrk (locale, "_.@+,");
+  language_end = strpbrk (locale, "_.@");
   if (language_end != NULL)
     {
       size_t len;
@@ -716,186 +791,11 @@ canonical_locale_charset ()
 static const char *
 englishname_of_language ()
 {
-  /* Derived from ISO 639.  */
-  static struct { const char *code; const char *english; } table[] = {
-    { "aa", "Afar" },
-    { "ab", "Abkhazian" },
-    { "ae", "Avestan" },
-    { "af", "Afrikaans" },
-    { "am", "Amharic" },
-    { "an", "Aragonese" },
-    { "ar", "Arabic" },
-    { "as", "Assamese" },
-    { "ay", "Aymara" },
-    { "az", "Azerbaijani" },
-    { "ba", "Bashkir" },
-    { "be", "Belarusian" },
-    { "bg", "Bulgarian" },
-    { "bh", "Bihari" },
-    { "bi", "Bislama" },
-    { "bn", "Bengali" },
-    { "bo", "Tibetan" },
-    { "br", "Breton" },
-    { "bs", "Bosnian" },
-    { "ca", "Catalan" },
-    { "ce", "Chechen" },
-    { "ch", "Chamorro" },
-    { "co", "Corsican" },
-    { "cs", "Czech" },
-    { "cu", "Church Slavic" },
-    { "cv", "Chuvash" },
-    { "cy", "Welsh" },
-    { "da", "Danish" },
-    { "de", "German" },
-    { "dz", "Dzongkha" },
-    { "el", "Greek" },
-    { "en", "English" },
-    { "eo", "Esperanto" },
-    { "es", "Spanish" },
-    { "et", "Estonian" },
-    { "eu", "Basque" },
-    { "fa", "Persian" },
-    { "fi", "Finnish" },
-    { "fj", "Fijian" },
-    { "fo", "Faroese" },
-    { "fr", "French" },
-    { "fy", "Frisian" },
-    { "ga", "Irish" },
-    { "gd", "Scots" },
-    { "gl", "Galician" },
-    { "gn", "Guarani" },
-    { "gu", "Gujarati" },
-    { "gv", "Manx" },
-    { "ha", "Hausa" },
-    { "he", "Hebrew" },
-    { "hi", "Hindi" },
-    { "ho", "Hiri Motu" },
-    { "hr", "Croatian" },
-    { "ht", "Haitian" },
-    { "hu", "Hungarian" },
-    { "hy", "Armenian" },
-    { "hz", "Herero" },
-    { "ia", "Interlingua" },
-    { "id", "Indonesian" },
-    { "ie", "Interlingue" },
-    { "ii", "Sichuan Yi" },
-    { "ik", "Inupiak" },
-    { "is", "Icelandic" },
-    { "it", "Italian" },
-    { "iu", "Inuktitut" },
-    { "ja", "Japanese" },
-    { "jw", "Javanese" },
-    { "ka", "Georgian" },
-    { "ki", "Kikuyu" },
-    { "kj", "Kuanyama" },
-    { "kk", "Kazakh" },
-    { "kl", "Kalaallisut" },
-    { "km", "Khmer" },
-    { "kn", "Kannada" },
-    { "ko", "Korean" },
-    { "ks", "Kashmiri" },
-    { "ku", "Kurdish" },
-    { "kv", "Komi" },
-    { "kw", "Cornish" },
-    { "ky", "Kirghiz" },
-    { "kok", "Konkani" },
-    { "la", "Latin" },
-    { "lb", "Letzeburgesch" },
-    { "li", "Limburgish" },
-    { "ln", "Lingala" },
-    { "lo", "Laotian" },
-    { "lt", "Lithuanian" },
-    { "lv", "Latvian" },
-    { "mg", "Malagasy" },
-    { "mh", "Marshall" },
-    { "mi", "Maori" },
-    { "mk", "Macedonian" },
-    { "ml", "Malayalam" },
-    { "mn", "Mongolian" },
-    { "mo", "Moldavian" },
-    { "mr", "Marathi" },
-    { "ms", "Malay" },
-    { "mt", "Maltese" },
-    { "my", "Burmese" },
-    { "mni", "Manipuri" },
-    { "na", "Nauru" },
-    { "nb", "Norwegian Bokmal" },
-    { "nd", "North Ndebele" },
-    { "ne", "Nepali" },
-    { "ng", "Ndonga" },
-    { "nl", "Dutch" },
-    { "nn", "Norwegian Nynorsk" },
-    { "no", "Norwegian" },
-    { "nr", "South Ndebele" },
-    { "nv", "Navajo" },
-    { "ny", "Nyanja" },
-    { "oc", "Occitan" },
-    { "om", "(Afan) Oromo" },
-    { "or", "Oriya" },
-    { "os", "Ossetian" },
-    { "pa", "Punjabi" },
-    { "pi", "Pali" },
-    { "pl", "Polish" },
-    { "ps", "Pashto" },
-    { "pt", "Portuguese" },
-    { "qu", "Quechua" },
-    { "rm", "Rhaeto-Roman" },
-    { "rn", "Kirundi" },
-    { "ro", "Romanian" },
-    { "ru", "Russian" },
-    { "rw", "Kinyarwanda" },
-    { "sa", "Sanskrit" },
-    { "sc", "Sardinian" },
-    { "sd", "Sindhi" },
-    { "se", "Northern Sami" },
-    { "sg", "Sango" },
-    { "si", "Sinhalese" },
-    { "sk", "Slovak" },
-    { "sl", "Slovenian" },
-    { "sm", "Samoan" },
-    { "sn", "Shona" },
-    { "so", "Somali" },
-    { "sq", "Albanian" },
-    { "sr", "Serbian" },
-    { "ss", "Siswati" },
-    { "st", "Sesotho" },
-    { "su", "Sundanese" },
-    { "sv", "Swedish" },
-    { "sw", "Swahili" },
-    { "ta", "Tamil" },
-    { "te", "Telugu" },
-    { "tg", "Tajik" },
-    { "th", "Thai" },
-    { "ti", "Tigrinya" },
-    { "tk", "Turkmen" },
-    { "tl", "Tagalog" },
-    { "tn", "Setswana" },
-    { "to", "Tonga" },
-    { "tr", "Turkish" },
-    { "ts", "Tsonga" },
-    { "tt", "Tatar" },
-    { "tw", "Twi" },
-    { "ty", "Tahitian" },
-    { "ug", "Uighur" },
-    { "uk", "Ukrainian" },
-    { "ur", "Urdu" },
-    { "uz", "Uzbek" },
-    { "vi", "Vietnamese" },
-    { "vo", "Volapuk" },
-    { "wo", "Wolof" },
-    { "wen", "Sorbian" },
-    { "xh", "Xhosa" },
-    { "yi", "Yiddish" },
-    { "yo", "Yoruba" },
-    { "za", "Zhuang" },
-    { "zh", "Chinese" },
-    { "zu", "Zulu" }
-  };
   size_t i;
 
-  for (i = 0; i < SIZEOF (table); i ++)
-    if (strcmp (table[i].code, language) == 0)
-      return table[i].english;
+  for (i = 0; i < language_table_size; i ++)
+    if (strcmp (language_table[i].code, language) == 0)
+      return language_table[i].english;
 
   return xasprintf ("Language %s", language);
 }
@@ -944,6 +844,7 @@ project_id ()
   if (linelen == (size_t)(-1))
     {
       error (0, 0, _("%s subprocess I/O error"), prog);
+      fclose (fp);
       goto failed;
     }
   if (linelen > 0 && line[linelen - 1] == '\n')
@@ -952,7 +853,7 @@ project_id ()
   fclose (fp);
 
   /* Remove zombie process from process list, and retrieve exit status.  */
-  exitstatus = wait_subprocess (child, prog, false, true, false);
+  exitstatus = wait_subprocess (child, prog, false, false, true, false);
   if (exitstatus != 0)
     {
       error (0, 0, _("%s subprocess failed with exit code %d"),
@@ -1011,6 +912,7 @@ project_id_version ()
   if (linelen == (size_t)(-1))
     {
       error (0, 0, _("%s subprocess I/O error"), prog);
+      fclose (fp);
       goto failed;
     }
   if (linelen > 0 && line[linelen - 1] == '\n')
@@ -1019,7 +921,7 @@ project_id_version ()
   fclose (fp);
 
   /* Remove zombie process from process list, and retrieve exit status.  */
-  exitstatus = wait_subprocess (child, prog, false, true, false);
+  exitstatus = wait_subprocess (child, prog, false, false, true, false);
   if (exitstatus != 0)
     {
       error (0, 0, _("%s subprocess failed with exit code %d"),
@@ -1146,7 +1048,7 @@ get_user_email ()
   /* Ask the user for his email address.  */
   argv[0] = "/bin/sh";
   argv[1] = (char *) prog;
-  argv[2] = _("\
+  argv[2] = (char *) _("\
 The new message catalog should contain your email address, so that users can\n\
 give you feedback about the translations, and so that maintainers can contact\n\
 you in case of unexpected technical problems.\n");
@@ -1169,6 +1071,7 @@ you in case of unexpected technical problems.\n");
   if (linelen == (size_t)(-1))
     {
       error (0, 0, _("%s subprocess I/O error"), prog);
+      fclose (fp);
       goto failed;
     }
   if (linelen > 0 && line[linelen - 1] == '\n')
@@ -1177,7 +1080,7 @@ you in case of unexpected technical problems.\n");
   fclose (fp);
 
   /* Remove zombie process from process list, and retrieve exit status.  */
-  exitstatus = wait_subprocess (child, prog, false, true, false);
+  exitstatus = wait_subprocess (child, prog, false, false, true, false);
   if (exitstatus != 0)
     {
       error (0, 0, _("%s subprocess failed with exit code %d"),
@@ -1256,7 +1159,7 @@ language_team_address ()
   fclose (fp);
 
   /* Remove zombie process from process list, and retrieve exit status.  */
-  exitstatus = wait_subprocess (child, prog, false, true, false);
+  exitstatus = wait_subprocess (child, prog, false, false, true, false);
   if (exitstatus != 0)
     {
       error (0, 0, _("%s subprocess failed with exit code %d"),
@@ -1313,12 +1216,12 @@ content_type (const char *header)
   old_field = get_field (header, "Content-Type");
   if (old_field != NULL)
     {
-      const char *charsetstr = strstr (old_field, "charset=");
+      const char *charsetstr = c_strstr (old_field, "charset=");
 
       if (charsetstr != NULL)
 	{
 	  charsetstr += strlen ("charset=");
-	  was_utf8 = (strcasecmp (charsetstr, "UTF-8") == 0);
+	  was_utf8 = (c_strcasecmp (charsetstr, "UTF-8") == 0);
 	}
     }
   return xasprintf ("text/plain; charset=%s",
@@ -1581,7 +1484,7 @@ subst_string (const char *str,
       size_t i;
       unsigned int j;
 
-      substlen = (size_t *) alloca (nsubst * sizeof (size_t));
+      substlen = (size_t *) xallocsa (nsubst * sizeof (size_t));
       for (j = 0; j < nsubst; j++)
 	{
 	  substlen[j] = strlen (subst[j][0]);
@@ -1613,6 +1516,8 @@ subst_string (const char *str,
 	  if (j == nsubst)
 	    i++;
 	}
+
+      freesa (substlen);
     }
 
   return str;
@@ -1652,13 +1557,10 @@ fill_header (msgdomain_list_ty *mdlp)
 	{
 	  message_ty *header_mp = NULL;
 	  char *header;
-	  const char *subst[4][2];
-	  const char *id;
-	  time_t now;
 
 	  /* Search the header entry.  */
 	  for (j = 0; j < mlp->nitems; j++)
-	    if (mlp->item[j]->msgid[0] == '\0' && !mlp->item[j]->obsolete)
+	    if (is_header (mlp->item[j]) && !mlp->item[j]->obsolete)
 	      {
 		header_mp = mlp->item[j];
 		break;
@@ -1669,7 +1571,7 @@ fill_header (msgdomain_list_ty *mdlp)
 	    {
 	      static lex_pos_ty pos = { __FILE__, __LINE__ };
 
-	      header_mp = message_alloc ("", NULL, "", 1, &pos);
+	      header_mp = message_alloc (NULL, "", NULL, "", 1, &pos);
 	      message_list_prepend (mlp, header_mp);
 	    }
 
@@ -1697,17 +1599,25 @@ fill_header (msgdomain_list_ty *mdlp)
 	  header_mp->msgstr_len = strlen (header) + 1;
 
 	  /* Update the comments in the header entry.  */
-	  id = project_id ();
-	  subst[0][0] = "SOME DESCRIPTIVE TITLE";
-	  subst[0][1] = xasprintf (get_title (), id, id);
-	  subst[1][0] = "PACKAGE";
-	  subst[1][1] = id;
-	  subst[2][0] = "FIRST AUTHOR <EMAIL@ADDRESS>";
-	  subst[2][1] = field_value[FIELD_LAST_TRANSLATOR];
-	  subst[3][0] = "YEAR";
-	  subst[3][1] =
-	    xasprintf ("%d", (time (&now), (localtime (&now))->tm_year + 1900));
-	  subst_string_list (header_mp->comment, SIZEOF (subst), subst);
+	  if (header_mp->comment != NULL)
+	    {
+	      const char *subst[4][2];
+	      const char *id;
+	      time_t now;
+
+	      id = project_id ();
+	      subst[0][0] = "SOME DESCRIPTIVE TITLE";
+	      subst[0][1] = xasprintf (get_title (), id, id);
+	      subst[1][0] = "PACKAGE";
+	      subst[1][1] = id;
+	      subst[2][0] = "FIRST AUTHOR <EMAIL@ADDRESS>";
+	      subst[2][1] = field_value[FIELD_LAST_TRANSLATOR];
+	      subst[3][0] = "YEAR";
+	      subst[3][1] =
+		xasprintf ("%d",
+			   (time (&now), (localtime (&now))->tm_year + 1900));
+	      subst_string_list (header_mp->comment, SIZEOF (subst), subst);
+	    }
 
 	  /* Finally remove the fuzzy attribute.  */
 	  header_mp->is_fuzzy = false;
@@ -1716,3 +1626,54 @@ fill_header (msgdomain_list_ty *mdlp)
 
   return mdlp;
 }
+
+
+/* Update the msgstr plural entries according to the nplurals count.  */
+static msgdomain_list_ty *
+update_msgstr_plurals (msgdomain_list_ty *mdlp)
+{
+  size_t k;
+
+  for (k = 0; k < mdlp->nitems; k++)
+    {
+      message_list_ty *mlp = mdlp->item[k]->messages;
+      message_ty *header_entry;
+      unsigned long int nplurals;
+      char *untranslated_plural_msgstr;
+      size_t j;
+
+      header_entry = message_list_search (mlp, NULL, "");
+      nplurals = get_plural_count (header_entry ? header_entry->msgstr : NULL);
+      untranslated_plural_msgstr = (char *) xmalloc (nplurals);
+      memset (untranslated_plural_msgstr, '\0', nplurals);
+
+      for (j = 0; j < mlp->nitems; j++)
+	{
+	  message_ty *mp = mlp->item[j];
+	  bool is_untranslated;
+	  const char *p;
+	  const char *pend;
+
+	  if (mp->msgid_plural != NULL)
+	    {
+	      /* Test if mp is untranslated.  (It most likely is.)  */
+	      is_untranslated = true;
+	      for (p = mp->msgstr, pend = p + mp->msgstr_len; p < pend; p++)
+		if (*p != '\0')
+		  {
+		    is_untranslated = false;
+		    break;
+		  }
+	      if (is_untranslated)
+		{
+		  /* Change mp->msgstr_len consecutive empty strings into
+		     nplurals consecutive empty strings.  */
+		  if (nplurals > mp->msgstr_len)
+		    mp->msgstr = untranslated_plural_msgstr;
+		  mp->msgstr_len = nplurals;
+		}
+	    }
+	}
+    }
+  return mdlp;
+}
