@@ -38,7 +38,7 @@
 #include "plural-exp.h"
 #include "plural-eval.h"
 #include "plural-table.h"
-#include "strstr.h"
+#include "c-strstr.h"
 #include "vasprintf.h"
 #include "exit.h"
 #include "message.h"
@@ -172,7 +172,7 @@ plural_help (const char *nullentry)
   const char *language;
   size_t j;
 
-  language = strstr (nullentry, "Language-Team: ");
+  language = c_strstr (nullentry, "Language-Team: ");
   if (language != NULL)
     {
       language += 15;
@@ -260,8 +260,8 @@ check_plural (message_list_ty *mlp, unsigned char **plural_distribution)
 
       nullentry = header->msgstr;
 
-      plural = strstr (nullentry, "plural=");
-      nplurals = strstr (nullentry, "nplurals=");
+      plural = c_strstr (nullentry, "plural=");
+      nplurals = c_strstr (nullentry, "nplurals=");
       if (plural == NULL && has_plural != NULL)
 	{
 	  const char *msg1 =
@@ -440,6 +440,9 @@ check_plural (message_list_ty *mlp, unsigned char **plural_distribution)
 static const message_ty *curr_mp;
 static lex_pos_ty curr_msgid_pos;
 static void
+formatstring_error_logger (const char *format, ...)
+     __attribute__ ((__format__ (__printf__, 1, 2)));
+static void
 formatstring_error_logger (const char *format, ...)
 {
   va_list args;
@@ -475,7 +478,6 @@ check_pair (const message_ty *mp,
   int seen_errors;
   int has_newline;
   unsigned int j;
-  const char *p;
 
   /* If the msgid string is empty we have the special entry reserved for
      information about the translation.  */
@@ -491,6 +493,8 @@ check_pair (const message_ty *mp,
 #define TEST_NEWLINE(p) (p[0] == '\n')
       if (msgid_plural != NULL)
 	{
+	  const char *p;
+
 	  if (TEST_NEWLINE(msgid_plural) != has_newline)
 	    {
 	      po_xerror (PO_SEVERITY_ERROR,
@@ -530,6 +534,8 @@ check_pair (const message_ty *mp,
 #define TEST_NEWLINE(p) (p[0] != '\0' && p[strlen (p) - 1] == '\n')
       if (msgid_plural != NULL)
 	{
+	  const char *p;
+
 	  if (TEST_NEWLINE(msgid_plural) != has_newline)
 	    {
 	      po_xerror (PO_SEVERITY_ERROR,
@@ -656,7 +662,7 @@ check_header_entry (const message_ty *mp, const char *msgstr_string)
 
   for (cnt = 0; cnt < nfields; ++cnt)
     {
-      char *endp = strstr (msgstr_string, required_fields[cnt]);
+      char *endp = c_strstr (msgstr_string, required_fields[cnt]);
 
       if (endp == NULL)
 	{
