@@ -1,5 +1,5 @@
 /* GNU gettext - internationalization aids
-   Copyright (C) 1995-1998, 2000-2005 Free Software Foundation, Inc.
+   Copyright (C) 1995-1998, 2000-2006 Free Software Foundation, Inc.
 
    This file was written by Peter Miller <millerp@canb.auug.org.au>
 
@@ -43,12 +43,9 @@
 #include "write-stringtable.h"
 #include "xalloc.h"
 #include "xallocsa.h"
-#include "strstr.h"
-#include "fwriteerror.h"
-#include "exit.h"
-#include "error-progname.h"
-#include "xerror.h"
-#include "po-error.h"
+#include "c-strstr.h"
+#include "xvasprintf.h"
+#include "po-xerror.h"
 #include "gettext.h"
 
 /* Our regular abbreviation.  */
@@ -318,28 +315,7 @@ message_print_comment_flags (const message_ty *mp, FILE *fp, bool debug)
 }
 
 
-/* =========== Some parameters for use by 'msgdomain_list_print'. ========== */
-
-
-/* This variable controls the page width when printing messages.
-   Defaults to PAGE_WIDTH if not set.  Zero (0) given to message_page_-
-   width_set will result in no wrapping being performed.  */
-static size_t page_width = PAGE_WIDTH;
-
-void
-message_page_width_set (size_t n)
-{
-  if (n == 0)
-    {
-      page_width = INT_MAX;
-      return;
-    }
-
-  if (n < 20)
-    n = 20;
-
-  page_width = n;
-}
+/* ========= Some parameters for use by 'msgdomain_list_print_po'. ========= */
 
 
 /* This variable controls the extent to which the page width applies.
@@ -379,27 +355,7 @@ message_print_style_escape (bool flag)
 }
 
 
-/* Whether to output a file in Java .properties syntax.  */
-static bool use_syntax_properties = false;
-
-void
-message_print_syntax_properties ()
-{
-  use_syntax_properties = true;
-}
-
-
-/* Whether to output a file in NeXTstep/GNUstep .strings syntax.  */
-static bool use_syntax_stringtable = false;
-
-void
-message_print_syntax_stringtable ()
-{
-  use_syntax_stringtable = true;
-}
-
-
-/* ================ msgdomain_list_print() and subroutines. ================ */
+/* =============== msgdomain_list_print_po() and subroutines. =============== */
 
 
 /* A version of memcpy optimized for the case n <= 1.  */
@@ -419,8 +375,10 @@ memcpy_small (void *dst, const void *src, size_t n)
 
 
 static void
-wrap (FILE *fp, const char *line_prefix, const char *name, const char *value,
-      enum is_wrap do_wrap, const char *charset)
+wrap (const message_ty *mp, FILE *fp, const char *line_prefix, int extra_indent,
+      const char *name, const char *value,
+      enum is_wrap do_wrap, size_t page_width,
+      const char *charset)
 {
   const char *canon_charset;
   const char *s;
@@ -487,12 +445,10 @@ wrap (FILE *fp, const char *line_prefix, const char *name, const char *value,
   first_line = true;
   do
     {
-      /* The \a and \v escapes were added by the ANSI C Standard.
-	 Prior to the Standard, most compilers did not have them.
-	 Because we need the same program on all platforms we don't provide
-	 support for them here.  Thus we only support \b\f\n\r\t.  */
+      /* The usual escapes, as defined by the ANSI C Standard.  */
 #     define is_escape(c) \
-       ((c) == '\b' || (c) == '\f' || (c) == '\n' || (c) == '\r' || (c) == '\t')
+        ((c) == '\a' || (c) == '\b' || (c) == '\f' || (c) == '\n' \
+         || (c) == '\r' || (c) == '\t' || (c) == '\v')
 
       const char *es;
       const char *ep;
@@ -552,7 +508,8 @@ wrap (FILE *fp, const char *line_prefix, const char *name, const char *value,
 		    {
 		      if (errno == EILSEQ)
 			{
-			  po_error (0, 0, _("invalid multibyte sequence"));
+			  po_xerror (PO_SEVERITY_ERROR, mp, NULL, 0, 0, false,
+				     _("invalid multibyte sequence"));
 			  continue;
 			}
 		      else
@@ -589,11 +546,13 @@ wrap (FILE *fp, const char *line_prefix, const char *name, const char *value,
 	    {
 	      switch (c)
 		{
+		case '\a': c = 'a'; break;
 		case '\b': c = 'b'; break;
 		case '\f': c = 'f'; break;
 		case '\n': c = 'n'; break;
 		case '\r': c = 'r'; break;
 		case '\t': c = 't'; break;
+		case '\v': c = 'v'; break;
 		default: abort ();
 		}
 	      *pp++ = '\\';
@@ -603,9 +562,15 @@ wrap (FILE *fp, const char *line_prefix, const char *name, const char *value,
 	      /* We warn about any use of escape sequences beside
 		 '\n' and '\t'.  */
 	      if (c != 'n' && c != 't')
-		po_error (0, 0, _("\
+		{
+		  char *error_message =
+		    xasprintf (_("\
 internationalized messages should not contain the `\\%c' escape sequence"),
-			  c);
+			       c);
+		  po_xerror (PO_SEVERITY_ERROR, mp, NULL, 0, 0, false,
+			     error_message);
+		  free (error_message);
+		}
 	    }
 	  else if (escape && !c_isprint ((unsigned char) c))
 	    {
@@ -658,7 +623,8 @@ internationalized messages should not contain the `\\%c' escape sequence"),
 		    {
 		      if (errno == EILSEQ)
 			{
-			  po_error (0, 0, _("invalid multibyte sequence"));
+			  po_xerror (PO_SEVERITY_ERROR, mp, NULL, 0, 0,
+				     false, _("invalid multibyte sequence"));
 			  continue;
 			}
 		      else
@@ -703,7 +669,7 @@ internationalized messages should not contain the `\\%c' escape sequence"),
 	 See INDENT-S.  */
       startcol_after_break = (line_prefix ? strlen (line_prefix) : 0);
       if (indent)
-	startcol_after_break = (startcol_after_break + 8) & ~7;
+	startcol_after_break = (startcol_after_break + extra_indent + 8) & ~7;
       startcol_after_break++;
 
       /* The line width.  Allow room for the closing quote character.  */
@@ -720,14 +686,14 @@ internationalized messages should not contain the `\\%c' escape sequence"),
 	{
 	  startcol += strlen (name);
 	  if (indent)
-	    startcol = (startcol + 8) & ~7;
+	    startcol = (startcol + extra_indent + 8) & ~7;
 	  else
 	    startcol++;
 	}
       else
 	{
 	  if (indent)
-	    startcol = (startcol + 8) & ~7;
+	    startcol = (startcol + extra_indent + 8) & ~7;
 	}
       /* Allow room for the opening quote character.  */
       startcol++;
@@ -764,13 +730,24 @@ internationalized messages should not contain the `\\%c' escape sequence"),
       if (first_line)
 	{
 	  fputs (name, fp);
-	  putc (indent ? '\t' : ' ', fp);
+	  if (indent)
+	    {
+	      if (extra_indent > 0)
+		fwrite ("        ", 1, extra_indent, fp);
+	      putc ('\t', fp);
+	    }
+	  else
+	    putc (' ', fp);
 	  first_line = false;
 	}
       else
 	{
 	  if (indent)
-	    putc ('\t', fp);
+	    {
+	      if (extra_indent > 0)
+		fwrite ("        ", 1, extra_indent, fp);
+	      putc ('\t', fp);
+	    }
 	}
 
       /* Print the portion itself, with linebreaks where necessary.  */
@@ -819,8 +796,10 @@ print_blank_line (FILE *fp)
 
 static void
 message_print (const message_ty *mp, FILE *fp, const char *charset,
-	       bool blank_line, bool debug)
+	       size_t page_width, bool blank_line, bool debug)
 {
+  int extra_indent;
+
   /* Separate messages with a blank line.  Uniforum doesn't like blank
      lines, so use an empty comment (unless there already is one).  */
   if (blank_line && (!uniforum
@@ -843,23 +822,61 @@ message_print (const message_ty *mp, FILE *fp, const char *charset,
   /* Print flag information in special comment.  */
   message_print_comment_flags (mp, fp, debug);
 
+  /* Print the previous msgid.  This helps the translator when the msgid has
+     only slightly changed.  */
+  if (mp->prev_msgctxt != NULL)
+    wrap (mp, fp, "#| ", 0, "msgctxt", mp->prev_msgctxt, mp->do_wrap,
+	  page_width, charset);
+  if (mp->prev_msgid != NULL)
+    wrap (mp, fp, "#| ", 0, "msgid", mp->prev_msgid, mp->do_wrap, page_width,
+	  charset);
+  if (mp->prev_msgid_plural != NULL)
+    wrap (mp, fp, "#| ", 0, "msgid_plural", mp->prev_msgid_plural, mp->do_wrap,
+	  page_width, charset);
+  extra_indent = (mp->prev_msgctxt != NULL || mp->prev_msgid != NULL
+		  || mp->prev_msgid_plural != NULL
+		  ? 3
+		  : 0);
+
   /* Print each of the message components.  Wrap them nicely so they
      are as readable as possible.  If there is no recorded msgstr for
      this domain, emit an empty string.  */
+  if (mp->msgctxt != NULL && !is_ascii_string (mp->msgctxt)
+      && po_charset_canonicalize (charset) != po_charset_utf8)
+    {
+      char *warning_message =
+	xasprintf (_("\
+The following msgctxt contains non-ASCII characters.\n\
+This will cause problems to translators who use a character encoding\n\
+different from yours. Consider using a pure ASCII msgctxt instead.\n\
+%s\n"), mp->msgctxt);
+      po_xerror (PO_SEVERITY_WARNING, mp, NULL, 0, 0, true, warning_message);
+      free (warning_message);
+    }
   if (!is_ascii_string (mp->msgid)
       && po_charset_canonicalize (charset) != po_charset_utf8)
-    po_multiline_warning (xasprintf (_("warning: ")),
-			  xasprintf (_("\
+    {
+      char *warning_message =
+	xasprintf (_("\
 The following msgid contains non-ASCII characters.\n\
 This will cause problems to translators who use a character encoding\n\
 different from yours. Consider using a pure ASCII msgid instead.\n\
-%s\n"), mp->msgid));
-  wrap (fp, NULL, "msgid", mp->msgid, mp->do_wrap, charset);
+%s\n"), mp->msgid);
+      po_xerror (PO_SEVERITY_WARNING, mp, NULL, 0, 0, true, warning_message);
+      free (warning_message);
+    }
+  if (mp->msgctxt != NULL)
+    wrap (mp, fp, NULL, extra_indent, "msgctxt", mp->msgctxt, mp->do_wrap,
+	  page_width, charset);
+  wrap (mp, fp, NULL, extra_indent, "msgid", mp->msgid, mp->do_wrap,
+	  page_width, charset);
   if (mp->msgid_plural != NULL)
-    wrap (fp, NULL, "msgid_plural", mp->msgid_plural, mp->do_wrap, charset);
+    wrap (mp, fp, NULL, extra_indent, "msgid_plural", mp->msgid_plural,
+	  mp->do_wrap, page_width, charset);
 
   if (mp->msgid_plural == NULL)
-    wrap (fp, NULL, "msgstr", mp->msgstr, mp->do_wrap, charset);
+    wrap (mp, fp, NULL, extra_indent, "msgstr", mp->msgstr, mp->do_wrap,
+	  page_width, charset);
   else
     {
       char prefix_buf[20];
@@ -871,7 +888,8 @@ different from yours. Consider using a pure ASCII msgid instead.\n\
 	   p += strlen (p) + 1, i++)
 	{
 	  sprintf (prefix_buf, "msgstr[%u]", i);
-	  wrap (fp, NULL, prefix_buf, p, mp->do_wrap, charset);
+	  wrap (mp, fp, NULL, extra_indent, prefix_buf, p, mp->do_wrap,
+		page_width, charset);
 	}
     }
 }
@@ -879,8 +897,10 @@ different from yours. Consider using a pure ASCII msgid instead.\n\
 
 static void
 message_print_obsolete (const message_ty *mp, FILE *fp, const char *charset,
-			bool blank_line)
+			size_t page_width, bool blank_line)
 {
+  int extra_indent;
+
   /* If msgstr is the empty string we print nothing.  */
   if (mp->msgstr[0] == '\0')
     return;
@@ -893,6 +913,12 @@ message_print_obsolete (const message_ty *mp, FILE *fp, const char *charset,
   /* Print translator comment if available.  */
   message_print_comment (mp, fp);
 
+  /* Print xgettext extracted comments (normally empty).  */
+  message_print_comment_dot (mp, fp);
+
+  /* Print the file position comments (normally empty).  */
+  message_print_comment_filepos (mp, fp, uniforum, page_width);
+
   /* Print flag information in special comment.  */
   if (mp->is_fuzzy)
     {
@@ -910,22 +936,60 @@ message_print_obsolete (const message_ty *mp, FILE *fp, const char *charset,
       putc ('\n', fp);
     }
 
+  /* Print the previous msgid.  This helps the translator when the msgid has
+     only slightly changed.  */
+  if (mp->prev_msgctxt != NULL)
+    wrap (mp, fp, "#~| ", 0, "msgctxt", mp->prev_msgctxt, mp->do_wrap,
+	  page_width, charset);
+  if (mp->prev_msgid != NULL)
+    wrap (mp, fp, "#~| ", 0, "msgid", mp->prev_msgid, mp->do_wrap, page_width,
+	  charset);
+  if (mp->prev_msgid_plural != NULL)
+    wrap (mp, fp, "#~| ", 0, "msgid_plural", mp->prev_msgid_plural, mp->do_wrap,
+	  page_width, charset);
+  extra_indent = (mp->prev_msgctxt != NULL || mp->prev_msgid != NULL
+		  || mp->prev_msgid_plural != NULL
+		  ? 1
+		  : 0);
+
   /* Print each of the message components.  Wrap them nicely so they
      are as readable as possible.  */
+  if (mp->msgctxt != NULL && !is_ascii_string (mp->msgctxt)
+      && po_charset_canonicalize (charset) != po_charset_utf8)
+    {
+      char *warning_message =
+	xasprintf (_("\
+The following msgctxt contains non-ASCII characters.\n\
+This will cause problems to translators who use a character encoding\n\
+different from yours. Consider using a pure ASCII msgctxt instead.\n\
+%s\n"), mp->msgctxt);
+      po_xerror (PO_SEVERITY_WARNING, mp, NULL, 0, 0, true, warning_message);
+      free (warning_message);
+    }
   if (!is_ascii_string (mp->msgid)
       && po_charset_canonicalize (charset) != po_charset_utf8)
-    po_multiline_warning (xasprintf (_("warning: ")),
-			  xasprintf (_("\
+    {
+      char *warning_message =
+	xasprintf (_("\
 The following msgid contains non-ASCII characters.\n\
 This will cause problems to translators who use a character encoding\n\
 different from yours. Consider using a pure ASCII msgid instead.\n\
-%s\n"), mp->msgid));
-  wrap (fp, "#~ ", "msgid", mp->msgid, mp->do_wrap, charset);
+%s\n"), mp->msgid);
+      po_xerror (PO_SEVERITY_WARNING, mp, NULL, 0, 0, true, warning_message);
+      free (warning_message);
+    }
+  if (mp->msgctxt != NULL)
+    wrap (mp, fp, "#~ ", extra_indent, "msgctxt", mp->msgctxt, mp->do_wrap,
+	  page_width, charset);
+  wrap (mp, fp, "#~ ", extra_indent, "msgid", mp->msgid, mp->do_wrap,
+	page_width, charset);
   if (mp->msgid_plural != NULL)
-    wrap (fp, "#~ ", "msgid_plural", mp->msgid_plural, mp->do_wrap, charset);
+    wrap (mp, fp, "#~ ", extra_indent, "msgid_plural", mp->msgid_plural,
+	  mp->do_wrap, page_width, charset);
 
   if (mp->msgid_plural == NULL)
-    wrap (fp, "#~ ", "msgstr", mp->msgstr, mp->do_wrap, charset);
+    wrap (mp, fp, "#~ ", extra_indent, "msgstr", mp->msgstr, mp->do_wrap,
+	  page_width, charset);
   else
     {
       char prefix_buf[20];
@@ -937,14 +1001,16 @@ different from yours. Consider using a pure ASCII msgid instead.\n\
 	   p += strlen (p) + 1, i++)
 	{
 	  sprintf (prefix_buf, "msgstr[%u]", i);
-	  wrap (fp, "#~ ", prefix_buf, p, mp->do_wrap, charset);
+	  wrap (mp, fp, "#~ ", extra_indent, prefix_buf, p, mp->do_wrap,
+		page_width, charset);
 	}
     }
 }
 
 
 static void
-msgdomain_list_print_po (msgdomain_list_ty *mdlp, FILE *fp, bool debug)
+msgdomain_list_print_po (msgdomain_list_ty *mdlp, FILE *fp, size_t page_width,
+			 bool debug)
 {
   size_t j, k;
   bool blank_line;
@@ -955,7 +1021,7 @@ msgdomain_list_print_po (msgdomain_list_ty *mdlp, FILE *fp, bool debug)
     {
       message_list_ty *mlp;
       const char *header;
-      char *charset;
+      const char *charset;
       char *allocated_charset;
 
       /* If the first domain is the default, don't bother emitting
@@ -974,7 +1040,7 @@ msgdomain_list_print_po (msgdomain_list_ty *mdlp, FILE *fp, bool debug)
       /* Search the header entry.  */
       header = NULL;
       for (j = 0; j < mlp->nitems; ++j)
-	if (mlp->item[j]->msgid[0] == '\0' && !mlp->item[j]->obsolete)
+	if (is_header (mlp->item[j]) && !mlp->item[j]->obsolete)
 	  {
 	    header = mlp->item[j]->msgstr;
 	    break;
@@ -985,7 +1051,7 @@ msgdomain_list_print_po (msgdomain_list_ty *mdlp, FILE *fp, bool debug)
       allocated_charset = NULL;
       if (header != NULL)
 	{
-	  const char *charsetstr = strstr (header, "charset=");
+	  const char *charsetstr = c_strstr (header, "charset=");
 
 	  if (charsetstr != NULL)
 	    {
@@ -993,9 +1059,10 @@ msgdomain_list_print_po (msgdomain_list_ty *mdlp, FILE *fp, bool debug)
 
 	      charsetstr += strlen ("charset=");
 	      len = strcspn (charsetstr, " \t\n");
-	      charset = allocated_charset = (char *) xallocsa (len + 1);
-	      memcpy (charset, charsetstr, len);
-	      charset[len] = '\0';
+	      allocated_charset = (char *) xallocsa (len + 1);
+	      memcpy (allocated_charset, charsetstr, len);
+	      allocated_charset[len] = '\0';
+	      charset = allocated_charset;
 
 	      /* Treat the dummy default value as if it were absent.  */
 	      if (strcmp (charset, "CHARSET") == 0)
@@ -1007,7 +1074,8 @@ msgdomain_list_print_po (msgdomain_list_ty *mdlp, FILE *fp, bool debug)
       for (j = 0; j < mlp->nitems; ++j)
 	if (!mlp->item[j]->obsolete)
 	  {
-	    message_print (mlp->item[j], fp, charset, blank_line, debug);
+	    message_print (mlp->item[j], fp, charset, page_width, blank_line,
+			   debug);
 	    blank_line = true;
 	  }
 
@@ -1015,7 +1083,8 @@ msgdomain_list_print_po (msgdomain_list_ty *mdlp, FILE *fp, bool debug)
       for (j = 0; j < mlp->nitems; ++j)
 	if (mlp->item[j]->obsolete)
 	  {
-	    message_print_obsolete (mlp->item[j], fp, charset, blank_line);
+	    message_print_obsolete (mlp->item[j], fp, charset, page_width,
+				    blank_line);
 	    blank_line = true;
 	  }
 
@@ -1025,226 +1094,14 @@ msgdomain_list_print_po (msgdomain_list_ty *mdlp, FILE *fp, bool debug)
 }
 
 
-void
-msgdomain_list_print (msgdomain_list_ty *mdlp, const char *filename,
-		      bool force, bool debug)
+/* Describes a PO file in .po syntax.  */
+const struct catalog_output_format output_format_po =
 {
-  FILE *fp;
-
-  /* We will not write anything if, for every domain, we have no message
-     or only the header entry.  */
-  if (!force)
-    {
-      bool found_nonempty = false;
-      size_t k;
-
-      for (k = 0; k < mdlp->nitems; k++)
-	{
-	  message_list_ty *mlp = mdlp->item[k]->messages;
-
-	  if (!(mlp->nitems == 0
-		|| (mlp->nitems == 1 && mlp->item[0]->msgid[0] == '\0')))
-	    {
-	      found_nonempty = true;
-	      break;
-	    }
-	}
-
-      if (!found_nonempty)
-	return;
-    }
-
-  /* Check whether the output format can accomodate all messages.  */
-  if (use_syntax_properties || use_syntax_stringtable)
-    {
-      if (mdlp->nitems > 1)
-	{
-	  if (use_syntax_properties)
-	    po_error (EXIT_FAILURE, 0, _("Cannot output multiple translation domains into a single file with Java .properties syntax. Try using PO file syntax instead."));
-	  if (use_syntax_stringtable)
-	    po_error (EXIT_FAILURE, 0, _("Cannot output multiple translation domains into a single file with NeXTstep/GNUstep .strings syntax."));
-	}
-      if (mdlp->nitems == 1)
-	{
-	  message_list_ty *mlp = mdlp->item[0]->messages;
-	  const lex_pos_ty *has_plural;
-	  size_t j;
-
-	  has_plural = NULL;
-	  for (j = 0; j < mlp->nitems; j++)
-	    {
-	      message_ty *mp = mlp->item[j];
-
-	      if (mp->msgid_plural != NULL)
-		{
-		  has_plural = &mp->pos;
-		  break;
-		}
-	    }
-
-	  if (has_plural != NULL)
-	    {
-	      error_with_progname = false;
-	      if (use_syntax_properties)
-		po_error_at_line (EXIT_FAILURE, 0,
-				  has_plural->file_name, has_plural->line_number,
-				  _("message catalog has plural form translations, but the output format does not support them. Try generating a Java class using \"msgfmt --java\", instead of a properties file."));
-	      if (use_syntax_stringtable)
-		po_error_at_line (EXIT_FAILURE, 0,
-				  has_plural->file_name, has_plural->line_number,
-				  _("message catalog has plural form translations, but the output format does not support them."));
-	      error_with_progname = true;
-	    }
-	}
-    }
-
-  /* Open the output file.  */
-  if (filename != NULL && strcmp (filename, "-") != 0
-      && strcmp (filename, "/dev/stdout") != 0)
-    {
-      fp = fopen (filename, "w");
-      if (fp == NULL)
-	po_error (EXIT_FAILURE, errno, _("cannot create output file \"%s\""),
-		  filename);
-    }
-  else
-    {
-      fp = stdout;
-      /* xgettext:no-c-format */
-      filename = _("standard output");
-    }
-
-  if (use_syntax_properties)
-    msgdomain_list_print_properties (mdlp, fp, page_width, debug);
-  else if (use_syntax_stringtable)
-    msgdomain_list_print_stringtable (mdlp, fp, page_width, debug);
-  else
-    msgdomain_list_print_po (mdlp, fp, debug);
-
-  /* Make sure nothing went wrong.  */
-  if (fwriteerror (fp))
-    po_error (EXIT_FAILURE, errno, _("error while writing \"%s\" file"),
-	      filename);
-}
-
-
-/* =============================== Sorting. ================================ */
-
-
-static int
-cmp_by_msgid (const void *va, const void *vb)
-{
-  const message_ty *a = *(const message_ty **) va;
-  const message_ty *b = *(const message_ty **) vb;
-  /* Because msgids normally contain only ASCII characters, it is OK to
-     sort them as if we were in the C locale. And strcoll() in the C locale
-     is the same as strcmp().  */
-  return strcmp (a->msgid, b->msgid);
-}
-
-
-void
-msgdomain_list_sort_by_msgid (msgdomain_list_ty *mdlp)
-{
-  size_t k;
-
-  for (k = 0; k < mdlp->nitems; k++)
-    {
-      message_list_ty *mlp = mdlp->item[k]->messages;
-
-      if (mlp->nitems > 0)
-	qsort (mlp->item, mlp->nitems, sizeof (mlp->item[0]), cmp_by_msgid);
-    }
-}
-
-
-/* Sort the file positions of every message.  */
-
-static int
-cmp_filepos (const void *va, const void *vb)
-{
-  const lex_pos_ty *a = (const lex_pos_ty *) va;
-  const lex_pos_ty *b = (const lex_pos_ty *) vb;
-  int cmp;
-
-  cmp = strcmp (a->file_name, b->file_name);
-  if (cmp == 0)
-    cmp = (int) a->line_number - (int) b->line_number;
-
-  return cmp;
-}
-
-static void
-msgdomain_list_sort_filepos (msgdomain_list_ty *mdlp)
-{
-  size_t j, k;
-
-  for (k = 0; k < mdlp->nitems; k++)
-    {
-      message_list_ty *mlp = mdlp->item[k]->messages;
-
-      for (j = 0; j < mlp->nitems; j++)
-	{
-	  message_ty *mp = mlp->item[j];
-
-	  if (mp->filepos_count > 0)
-	    qsort (mp->filepos, mp->filepos_count, sizeof (mp->filepos[0]),
-		   cmp_filepos);
-	}
-    }
-}
-
-
-/* Sort the messages according to the file position.  */
-
-static int
-cmp_by_filepos (const void *va, const void *vb)
-{
-  const message_ty *a = *(const message_ty **) va;
-  const message_ty *b = *(const message_ty **) vb;
-  int cmp;
-
-  /* No filepos is smaller than any other filepos.  */
-  if (a->filepos_count == 0)
-    {
-      if (b->filepos_count != 0)
-	return -1;
-    }
-  if (b->filepos_count == 0)
-    return 1;
-
-  /* Compare on the file names...  */
-  cmp = strcmp (a->filepos[0].file_name, b->filepos[0].file_name);
-  if (cmp != 0)
-    return cmp;
-
-  /* If they are equal, compare on the line numbers...  */
-  cmp = a->filepos[0].line_number - b->filepos[0].line_number;
-  if (cmp != 0)
-    return cmp;
-
-  /* If they are equal, compare on the msgid strings.  */
-  /* Because msgids normally contain only ASCII characters, it is OK to
-     sort them as if we were in the C locale. And strcoll() in the C locale
-     is the same as strcmp().  */
-  return strcmp (a->msgid, b->msgid);
-}
-
-
-void
-msgdomain_list_sort_by_filepos (msgdomain_list_ty *mdlp)
-{
-  size_t k;
-
-  /* It makes sense to compare filepos[0] of different messages only after
-     the filepos[] array of each message has been sorted.  Sort it now.  */
-  msgdomain_list_sort_filepos (mdlp);
-
-  for (k = 0; k < mdlp->nitems; k++)
-    {
-      message_list_ty *mlp = mdlp->item[k]->messages;
-
-      if (mlp->nitems > 0)
-	qsort (mlp->item, mlp->nitems, sizeof (mlp->item[0]), cmp_by_filepos);
-    }
-}
+  msgdomain_list_print_po,		/* print */
+  false,				/* requires_utf8 */
+  true,					/* supports_multiple_domains */
+  true,					/* supports_contexts */
+  true,					/* supports_plurals */
+  false,				/* alternative_is_po */
+  false					/* alternative_is_java_class */
+};
